package Imager::Font::Type1;
use 5.006;
use strict;
use Imager::Font::T1;

our @ISA = qw(Imager::Font::T1);

our $VERSION = "1.013";

1;

__END__

=head1 NAME

  Imager::Font::Type1 - low-level functions for T1Lib text output

=head1 DESCRIPTION

This is a simple wrapper around Imager::Font::T1 for backwards
compatibility.

=head1 AUTHOR

Tony Cook <tonyc@cpan.org>

=cut
